#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

_default_template = """

using namespace std;

#include <iostream>
#include "iterators.cpp"
template<typename T>
void Print(T x) {
    std::cout << x;
}

void Print() {
    std::cout << std::endl;
}
void DebugFunction() {
    std::cout << "DEBUG FUNCTION CALLED" << std::endl;
}

int main(int argc, char *argv[])
{
%(FRAME_TEXT)s

return 0;
}
"""

_arduino_template = """
#include "iterators.cpp"

void setup()
{
%(FRAME_TEXT)s
}

void loop()
{
}
"""


_default_makefile_template = """
all :
\tg++ %(filename)s.c -o %(filename)s
"""

_arduino_makefile_template = """
BOARD_TAG    = leonardo
ARDUINO_PORT = /dev/ttyACM0
CPPFLAGS = -std=c++11

include /usr/share/arduino/Arduino.mk
"""

def default_result_filename(build_dir, cname):
    result = os.path.join(build_dir, cname)
    return result

def arduino_result_filename(build_dir, cname):
    hexfile = None
    cbuild_dir = os.path.join(build_dir, "build-leonardo")

    for filename in os.listdir(cbuild_dir):
        if filename.endswith(".hex"):
            hexfile = filename
            break

    if hexfile:
        return os.path.join(cbuild_dir, hexfile)

    # raise Exception("Not Finished")


cpp_templates = {
    "default" : _default_template,
    "arduino" : _arduino_template,
}

makefile_templates = {
    "default" : _default_makefile_template,
    "arduino" : _arduino_makefile_template,
}

mainfile_extensions = {
    "default" : "c",
    "arduino" : "ino",
}

clib_exclusions = {
    "default" : [],
    "arduino" : [],
}

result_file = {
    "default" : default_result_filename,
    "arduino" : arduino_result_filename,

}

modify_result_file = {
    "default": lambda x: x,
    "arduino": lambda x: x+".hex",
}



#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

#
# I was going to have a stack of contexts. Instead, I'll maintain
# a stack but allow them to point to their parent. This actually means
# all contexts actually form a graph - which is true. (Many local contexts can
# point at a parent context)
#
# This then allows looking in the current context, and then outwards.
#
# A context is a collection of names inside a scope.
#
# Names that are bound within that scope can be well defined in that they bind to a
# literal value that has a type, or to the result of an expression which again has
# a clear derivable state.
#
# Names that are not bound into this state, by contrast are free variables.
# Free variables are not allows in little python at present, since they are potentially
# poly-typed
#
# While I can see my original point, having PyIdentifiers as the values here is kinda odd,
# since really the values associated with the names is the key point here, and the values
# are themselves more expression-result-types, not identifiers.
#
# As a result, contexts themselves need rework here
#
# Context Leakage is a very bad name.
#
# In fact, in any given context, an identifier gets assigned the results of an expression
# This may happen more than once.
#
# There are some problem scenarios:
#
#  * The results of two (or more) expressions assigned to a name are conflicting types
#  * An expression uses a name that is not yet well defined at that point in a program
#
# Suggests the following strategy:
#
#  * So, we should go through the program, assigning *expressions* to names in contexts
#  * When we need the type of an identifier, we look in it's context for the name, find
#    an expression and evaluate it's type

#

class UndefinedValue(Exception):
    pass

class Context(object):

    contexts = {}
    def __init__(self, parent=None):
        self.names = {} # Stores the type against the name.
        self.parent = parent
        if id(self) not in self.contexts:
            self.contexts[id(self)] = self

    def store(self, name, expression):
        print "Context.store NAME", name, "VALUE", expression
        if name in self.names:
            print "WARNING: Name %s already exists in names, this may be OK. Storing expression %s" % (repr(name), repr(expression))
        try:
            self.names[name].append( expression )
        except KeyError:
            self.names[name] = [ expression ]
        print "CONTEXT", self.names

    def lookup(self, name):
        if name in self.names:
            values = self.names[name]
            return values[0]
        if self.parent:
            values = self.parent.lookup(name)
            return values[0]
        raise UndefinedValue("Cannot find name %s in current context stack" % name)

    def __repr__(self):
        return repr(id(self))

    def __json__(self):
        names_info = {}
        for name in self.names:
            if type(self.names[name]) == type(""):
                names_info[name] = self.names[name]
            else:
                names_info[name] = self.names[name] # .__info__()
        return {"id": id(self), "parent": id(self.parent) if self.parent is not None else None, "names": names_info}




#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Tree node to be used by the pynodes. Possibly other nodes later.
# Added once the pynode parsing required it.
#

def cmp_first(x,y):
    hx,tx = x
    hy,ty = y
    return cmp(hx,hy)

class Tree(object):
    def __init__(self):
        self.children = [] # Children are ordered

    def __pson__(self):
        return { "nId" : str(id(self)),
                 "nStr" : str(self)[:40],
                 "nchildren" : [x.__pson__() for x in self.children] }
    #
    def add_child(self, node):
        self.children.append(node)
    #
    def add_children(self, *nodes):
        self.children = self.children + list(nodes)
    #
    def depth_walk(self):
        "Return a generator that is a depth first walk of the tree"
        for node in self.children:
            for child in node.depth_walk():
                yield child
        yield self

    def _breadth_walk_children(self,depth=2):
        nodes = []
        for node in self.children:
            nodes.append( (depth, node ) )

        for node in self.children:
            for cdepth, child in node._breadth_walk_children(depth+1):
                nodes.append( (cdepth, child ) )

        nodes.sort(cmp_first)
        for node in nodes:
            yield node

    def breadth_walk(self):
        "Return a generator that is a breadth first walk of the tree, not optimised, copes with unbalanced trees"
        yield self
        nodes = []
        for node in self._breadth_walk_children():
            nodes.append(node)

        nodes.sort(cmp_first)
        for depth, node in nodes:
            yield node

    def breadth_walk_up(self):
        "Return a generator that is a breadth first walk of the tree. Starts with lowest layers. Not optimised, copes with unbalanced trees"
        nodes = []
        for node in self._breadth_walk_children():
            nodes.append(node)

        nodes.sort(cmp_first)
        max_depth = 0
        for depth, node in nodes:
            if depth> max_depth:
                max_depth=depth

        for this_depth in range(max_depth,1,-1):
            for depth, node in nodes:
                if depth == this_depth:
                    yield node
        yield self

def _run_tests():
    "Basic smoke tests for the tree type"
    class Node(Tree):
        def __init__(self, tag):
            self.tag = tag
            super(Node, self).__init__()
        def info(self):
            return {"tag" : self.tag }

        def DF_walk(self):
            "Process the tree depth first"
            for i in self.depth_walk():
                print "DFS", i

        def BF_walk(self):
            "Process the tree breadth first"
            for i in a.breadth_walk():
                print "BFS", i

        def BFUP_walk(self):
            "Process the tree breadth first"
            for i in a.breadth_walk_up():
                print "BFUP", i

        def __repr__(self):
            return "N:" + str(self.tag)

    # Build the following tree:
    #
    #                 A1                        1
    #       B2                  C3              2
    #   D4       E5        F6        G7         3
    #  H8 I9   J10 K11   L12 M13   N14 O15      4
    # P16                                Q17     5
    #    R18                                S19  6
    #
    a = Node(1)
    b = Node(2)
    c = Node(3)
    d = Node(4)
    e = Node(5)
    f = Node(6)
    g = Node(7)

    h = Node(8)
    i = Node(9)
    j = Node(10)
    k = Node(11)
    l = Node(12)
    m = Node(13)
    n = Node(14)
    o = Node(15)

    p = Node(16)
    q = Node(17)
    r = Node(18)
    s = Node(19)

    h.add_children(p)
    p.add_children(r)
    o.add_children(q)
    q.add_children(s)


    a.add_children(b,c)
    b.add_children(d,e)
    c.add_children(f,g)
    d.add_children(h,i)
    e.add_children(j,k)
    f.add_children(l,m)
    g.add_children(n,o)

    print "df_walk"
    a.DF_walk()

    print "bf_walk"
    a.BF_walk()

    print "bfup_walk"
    a.BFUP_walk()



if __name__ == "__main__":
    _run_tests()



#!/usr/bin/python

#
# Note: This file is autogenerated from clib, using clib/mk_py_clib.py and the C++ files.
# Make edits there not here - changes here will be overwritten!
#
files = {'iterators.cpp': '#include "iterators.hpp"\n\nstruct range : public Generator<int> {\n    int start;\n    int end;\n    int step;\n\n    int index;\n\n    range() :                                start(0),     end(0),   step(1), index(0)            {     };\n    range(int end) :                         start(0),     end(end), step(1), index(0)            {     };\n    range(int start, int end)     :          start(start), end(end), step(1), index(start)        {     };\n    range(int start, int end, int stepsize): start(start), end(end), step(stepsize), index(start) {     };\n    ~range() {     };\n\n    virtual int next() {\n        GENERATOR_START\n\n        while ( step>0  ? index < end : index > end) {\n            YIELD(index);\n            index = index + step;\n        }\n\n        GENERATOR_END\n    }\n};\n', 'iterators.hpp': '#ifndef PYXIE_ITERATORS_HPP\n#define PYXIE_ITERATORS_HPP\n\n/*\n * Python Style Iterators in C++, based in part on experiment work in Kamaelia\n * Redone to remove use of generators, so that this can be used on Arduino\n * \n */\n\n#define GENERATOR_START if (this->__generator_state == -1) { return __default_value; } switch(this->__generator_state) { default:\n#define YIELD(value)    this->__generator_state = __LINE__; return ((value) );   case __LINE__:\n#define GENERATOR_END    }; this->__generator_state = -1; return __default_value;\n\ntemplate<class T>\nstruct Iterator {\n    virtual T next()=0;\n    virtual bool completed()=0;\n};\n\ntemplate<class T>\nclass Generator : public Iterator<T> {\n  protected:\n    int __generator_state;\n  public:\n    T __default_value;\n    Generator() : __default_value(T()) {     };\n    ~Generator() {     };\n    virtual bool completed() { return __generator_state==-1; };\n};\n\n#endif\n', 'iterators_test.cpp': '/*\n * Test of a C++ version of Python style generators.\n * \n */\n\n#include <iostream>\n#include "iterators.cpp"\n/*\n\nWhile it may be a little unclear, this C++ program is equivalent\nto this python program:\n\nfor count in range(5):\n    print count,\n\nprint\n\nIt doesn\'t do this:\n\nfor(int count=0; count<5; count++) {\n    std::cout << count;\n}\nstd::cout << "." << std::endl;\n\nBecause doing it the way we do makes it avoid treating for/range as a\nspecial case, meaning we solve the harder problem first.  Optimisations\nspecific to certain code structures can come later.\n\n*/\n\n\nint main(int argc, char* argv[]) {\n    // Code to be replaced with a range() style iterator\n    int count;\n    range range_gen = range(5);\n\n    while (true) {\n        count = range_gen.next();\n        if (range_gen.completed())\n            break;\n        std::cout << count;\n    }\n    std::cout << "." << std::endl;\n\n    return 0;\n}\n'}




#include <Servo.h>

# For testing purposes...
def trace(func, args):
    print func, args

HIGH="HIGH"
LOW="LOW"
OUTPUT="OUTPUT"

def digitalWrite(*args):         trace("digitalWrite", args)
def delayMicroseconds(*args):    trace("delayMicroseconds", args)
def pinMode(*args):              trace("pinMode", args)
def analogRead(*args):           trace("analogRead", args); return 0
def millis(*args):               trace("millis", args); return 0

class Servo(object):
    def __init__(self):
        self.pin = None
    def writeMicroseconds(self, number):
        if self.pin != None:
            trace("Servo.writeMicroseconds", [self.pin, number])
        else:
            trace("Servo.writeMicroseconds", number)
    def attach(self, pin):
        trace("Servo.attach", pin)
        self.pin = pin

# ::::::::::::::
# InternetLinks.ino
# ::::::::::::::
# Arduino Software:      http://arduino.cc/en/Main/Software
# USB Drivers:           http://www.silabs.com/products/mcu/Pages/USBtoUARTBridgeVCPDrivers.aspx
# DAGU Product Support:  https://sites.google.com/site/daguproducts/

# ::::::::::::::
# IOpins.h
# ::::::::::::::

FLHpin = 6      # Front Left  Hip        - digital output   D6
FRHpin = 11     # Front Right Hip        - digital output  D11
RLHpin = 5      # Rear  Left  Hip        - digital output   D5
RRHpin = 4      # Rear  Right Hip        - digital output   D4

FLKpin = 12     # Front Left  Knee       - digital output  D12
FRKpin = 13     # Front Right Knee       - digital output  D13
RLKpin = 2      # Rear  Left  Knee       - digital output   D2
RRKpin = 3      # Rear  Right Knee       - digital output   D3

PANpin = 14     # Neck  Pan   A0         - digital output  D14
TILpin = 15     # Neck  Tilt  A1         - digital output  D15

IRleft = 2      # Compound Eye Left      - analog  input    A2
IRright = 4     # Compound Eye Right     - analog  input    A3
IRup = 3        # Compound Eye Up        - analog  input    A4
IRdown = 5      # Compound Eye Down      - analog  input    A5
IRleds = 8      # Compound Eye LEDs      - digital output   D8

# ::::::::::::::
# Constants.h
# ::::::::::::::
Speed = 40           # time in mS allowed for each cycle for walk - must allow for servo speed
Stepsize = 250       # size of steps
Boredom = 6000       # time in milliseconds before robot gets bored and sits down

LRscalefactor = 2    # LR & UD scalefactors used to adjust pan/tilt servo responsiveness
UDscalefactor = 2    # smaller values will prevent overshoot but can make servos slow to respond
distancemax = 200    # distance values smaller than this are assumed out of range
bestdistance = 500   # robot will try to maintain this distance from the object

FLHcenter = 1500     # Front Left  Hip  center position - ajust to suit servos
FRHcenter = 1500     # Front Righ  Hip  center position - ajust to suit servos
RLHcenter = 1500     # Rear  Left  Hip  center position - ajust to suit servos
RRHcenter = 1500     # Rear  Right Hip  center position - ajust to suit servos

FLKcenter = 1500     # Front Left  Knee center position - ajust to suit servos
FRKcenter = 1500     # Front Righ  Knee center position - ajust to suit servos
RLKcenter = 1500     # Rear  Left  Knee center position - ajust to suit servos
RRKcenter = 1500     # Rear  Right Knee center position - ajust to suit servos

PANcenter = 1500     # Pan  servo center position
TILcenter = 1500     # Tilt servo center position

LRmax = 2300         # Pan  center position +800 - head turned 80 degrees  left
LRmin = 700          # Pan  center position -800 - head turned 80 degrees right
UDmax = 2000         # Look up   - adjust so head does not try to bend backward
UDmin = 700          # look down - ajust so head looks down without straining servo

IRdelay = 200        #  microseconds to wait while phototransistors respond to changes in IR

# ::::::::::::::
# PlayfulPuppy.ino
# ::::::::::::::


# -------------------------------------------------------------- define global variables and servos here ---------------------------
time = 0l           # "l" to hint 'unsigned long'  // timer used to control walk speed
sit = 0l            # "l" to hint 'unsigned long'  // timer used to determine lazyness and control sit movement
trick = 0l          # "l" to hint 'unsigned long'  // timer used to control jump, shake hand and lie down movement

LShift = 0          # int    // Speed and direction of left legs
RShift = 0          # int    // Speed and direction of right legs
Raise = 0           # int    // amount to lift legs by
cycle = 0           # int    // counter to control leg movements

pan=PANcenter       # int    // position of pan  servo - start at center
tilt=TILcenter      # int    // position of tilt servo - start at center

distance = 0        # int    // aproximate distance of object (average of sensor readings)
temp = 0L           # long   // tempory value used in various calculations

leftIRvalue = 0     # int   // Left  sensor value after correcting for ambient light
rightIRvalue = 0    # int   // Right sensor value after correcting for ambient light
upIRvalue = 0       # int   // Upper sensor value after correcting for ambient light
downIRvalue = 0     # int   // Lower sensor value after correcting for ambient light

# define hip servos
FLHservo = Servo()  # Servo FLHservo
FRHservo = Servo()  # Servo FRHservo
RLHservo = Servo()  # Servo RLHservo
RRHservo = Servo()  # Servo RRHservo

# define knee servos
FLKservo = Servo()  # Servo FLKservo
FRKservo = Servo()  # Servo FRKservo
RLKservo = Servo()  # Servo RLKservo
RRKservo = Servo()  # Servo RRKservo

# define pan, tilt and tail wag servos
PANservo = Servo()  # Servo PANservo
TILservo = Servo()  # Servo TILservo




#setup()
# def setup():
FLHservo.attach(FLHpin)                                    # attach hip  servos
FRHservo.attach(FRHpin)
RLHservo.attach(RLHpin)
RRHservo.attach(RRHpin)

FLKservo.attach(FLKpin)                                    # attach knee servos
FRKservo.attach(FRKpin)
RLKservo.attach(RLKpin)
RRKservo.attach(RRKpin)

PANservo.attach(PANpin)                                    # attach head servos
TILservo.attach(TILpin)

FLHservo.writeMicroseconds(FLHcenter)                      # Center servos
FRHservo.writeMicroseconds(FRHcenter)
RLHservo.writeMicroseconds(RLHcenter)
RRHservo.writeMicroseconds(RRHcenter)
FLKservo.writeMicroseconds(FLKcenter)
FRKservo.writeMicroseconds(FRKcenter)
RLKservo.writeMicroseconds(RLKcenter)
RRKservo.writeMicroseconds(RRKcenter)
PANservo.writeMicroseconds(PANcenter)
TILservo.writeMicroseconds(TILcenter)

pinMode(IRleds,OUTPUT)                                     # enable digital output to control IR leds on compound eye

while True:
    # loop()
    # ============================================================== Main Loop ==========================================================
    # def loop():
    # return;                                    # enable to freeze robot when setting servo center positions
    # IReye()                                                    # read eye sensors allowing for ambient light
    # ::::::::::::::
    # IReye.ino
    # ::::::::::::::
    # def IReye(): # ================================================== Read IR compound eye =========================
    #
    digitalWrite(IRleds,HIGH)                                  #  turn on IR LEDs to read TOTAL IR LIGHT (ambient + reflected)
    delayMicroseconds(IRdelay)                                 # allow time for phototransistors to respond.
    leftIRvalue  = analogRead(IRleft)                          # TOTAL IR = AMBIENT IR + LED IR REFLECTED FROM OBJECT
    rightIRvalue = analogRead(IRright)                         # TOTAL IR = AMBIENT IR + LED IR REFLECTED FROM OBJECT
    upIRvalue    = analogRead(IRup)                            # TOTAL IR = AMBIENT IR + LED IR REFLECTED FROM OBJECT
    downIRvalue  = analogRead(IRdown)                          # TOTAL IR = AMBIENT IR + LED IR REFLECTED FROM OBJECT
    #
    digitalWrite(IRleds, LOW)                                  # turn off IR LEDs to read AMBIENT IR LIGHT (IR from indoor lighting and sunlight)
    delayMicroseconds(IRdelay)                                 # allow time for phototransistors to respond.
    leftIRvalue  = leftIRvalue - analogRead(IRleft)            # REFLECTED IR = TOTAL IR - AMBIENT IR
    rightIRvalue = rightIRvalue - analogRead(IRright)          # REFLECTED IR = TOTAL IR - AMBIENT IR
    upIRvalue    = upIRvalue - analogRead(IRup)                # REFLECTED IR = TOTAL IR - AMBIENT IR
    downIRvalue  = downIRvalue - analogRead(IRdown)            # REFLECTED IR = TOTAL IR - AMBIENT IR
    #
    distance = (leftIRvalue + rightIRvalue + upIRvalue + downIRvalue) / 4 # distance of object is mean of reflected IR


    #IRtrack()                                                  # track objects with head
    # ::::::::::::::
    # IRtrack.ino
    # ::::::::::::::
    #def IRtrack(): # =============================================== Track object in range ==========================
    # global pan
    # global tilt
    if distance < distancemax:                          #  nothing in range - return head to center position
        if pan>PANcenter:
            pan = pan - 5
        if pan<PANcenter:
            pan = pan + 5
        if tilt>TILcenter:
            tilt = tilt - 3
        if tilt < TILcenter:
            tilt = tilt + 3
    else:
        # ---------------------------------------------------------- Track object with head --------------------
        # Consider meaning of "/" in embedded context
        panscale = (leftIRvalue + rightIRvalue) / LRscalefactor  # panscale & tiltscale used to adjust servo response according to distance
        tiltscale = (upIRvalue + downIRvalue) / UDscalefactor    # adjust LRscalefactor and UDscalefactor to tweak servo responsiveness
        #
        leftright = (rightIRvalue - leftIRvalue) * 5 / panscale
        pan = pan + leftright
        updown = (downIRvalue - upIRvalue) * 5 / tiltscale
        tilt = tilt + updown
        # 
        if pan < LRmin: pan=LRmin
        if pan > LRmax: pan=LRmax
        if tilt < UDmin: tilt=UDmin
        if tilt > UDmax: tilt=UDmax
    #
    # 
    # PANservo.writeMicroseconds(pan)
    # TILservo.writeMicroseconds(tilt)
    PANservo.writeMicroseconds(1500-(pan-1500))                 #  reverses direction of pan/tilt servos
    TILservo.writeMicroseconds(1500-(tilt-1500))


    # ------------------------------------------------------------ Decide what to do --------------------------------------------------
    if millis() - sit < Boredom:       # puppy is playing
        # Follow()                                      # follow object with body to keep in range
        # ::::::::::::::
        # Follow.ino
        # ::::::::::::::
        # def Follow(): # ================================================= Body follows object ============================
        # global LShift
        # global RShift
        LShift = 0                                                   # reset left and right speeds
        RShift = 0
        if not (distance<distancemax):   # nothing to follow - stop moving

            # ----------------------------------------Turn body to follow object if head turns too far ------------------
            temp = pan - PANcenter 
            if abs(temp) > 300:                                        # if the head turns off center by more than 300
              LShift = temp/2                                            # adjust left and right speeds to turn towards the object
              RShift = -temp/2

            # ------------------------------------------------------------ Move forward or backward to follow object ---------------------------------------
            temp = (distance - bestdistance) * 15 / 10                     # *15/10 = *1.5 but with whole numbers
            LShift = LShift - temp
            RShift = RShift - temp

            if(LShift>Stepsize):
                LShift=Stepsize

            if(LShift<-Stepsize):
                LShift=-Stepsize

            if(RShift>Stepsize):
                RShift=Stepsize

            if(RShift<-Stepsize):
                RShift=-Stepsize

        # Play()                                        # follows object or jump
        # ::::::::::::::
        # Play.ino
        # ::::::::::::::
        # def Play():
        # global LShift
        # global RShift
        # global time
        Raise = 150 +((abs(LShift) + abs(RShift)) /2)      # adjust hight of step to suit stride
        if Raise < 200:                                    # not worth moving - stand still
            LShift = 0
            RShift = 0
            Raise = 0
        else:
            sit = millis()                                 # reset sit timer if moving
            trick = millis()                               # reset trick timer

        if tilt < 980:                                     # if looking straight up then jump!
            # Jump()
            # ::::::::::::::
            # Jump.ino
            # ::::::::::::::
            # ============================================================ Jump onto rear legs ==============================================
            # def Jump():

            pan = PANcenter
            tilt = 1700
            # ------------------------------------------------------------ Lean forward -------------------------------------
            PANservo.writeMicroseconds(1500-(pan-1500))                #  move head into correct position
            TILservo.writeMicroseconds(1500-(tilt-1500))               #  head looks down when on hind legs
            RLHservo.writeMicroseconds(RLHcenter-600)                  # put rear hips backward to shift balance
            RRHservo.writeMicroseconds(RRHcenter+600) 
            FLKservo.writeMicroseconds(FLKcenter-600)                  #  straighten front knees in preparation to jump
            FRKservo.writeMicroseconds(FRKcenter-600) 
            delay(350)                                                 # wait for body to fall fall forward
            #
            # ------------------------------------------------------------ Jump up on back legs ----------------------------------
            FLHservo.writeMicroseconds(FLHcenter+600)                  # front hips forward
            FRHservo.writeMicroseconds(FRHcenter-600)
            RLHservo.writeMicroseconds(RLHcenter+600)                  # rear hips forward
            RRHservo.writeMicroseconds(RRHcenter-600)
            FLKservo.writeMicroseconds(FLKcenter+300)                  # front knees bent inward to push up
            FRKservo.writeMicroseconds(FRKcenter+300)
            RLKservo.writeMicroseconds(RLKcenter+100)                  # Bend knees slightly to create pivot point
            RRKservo.writeMicroseconds(RRKcenter+100)

            trick = millis()                                           # reset trick timer
            while millis()-trick<2000:                                  # minimum time to stay on back legs
                # IReye()                                                  # read eye sensors allowing for ambient light
                # ::::::::::::::
                # IReye.ino
                # ::::::::::::::
                # def IReye(): # ================================================== Read IR compound eye =========================
                #
                digitalWrite(IRleds,HIGH)                                  #  turn on IR LEDs to read TOTAL IR LIGHT (ambient + reflected)
                delayMicroseconds(IRdelay)                                 # allow time for phototransistors to respond.
                leftIRvalue  = analogRead(IRleft)                          # TOTAL IR = AMBIENT IR + LED IR REFLECTED FROM OBJECT
                rightIRvalue = analogRead(IRright)                         # TOTAL IR = AMBIENT IR + LED IR REFLECTED FROM OBJECT
                upIRvalue    = analogRead(IRup)                            # TOTAL IR = AMBIENT IR + LED IR REFLECTED FROM OBJECT
                downIRvalue  = analogRead(IRdown)                          # TOTAL IR = AMBIENT IR + LED IR REFLECTED FROM OBJECT
                #
                digitalWrite(IRleds, LOW)                                  # turn off IR LEDs to read AMBIENT IR LIGHT (IR from indoor lighting and sunlight)
                delayMicroseconds(IRdelay)                                 # allow time for phototransistors to respond.
                leftIRvalue  = leftIRvalue - analogRead(IRleft)            # REFLECTED IR = TOTAL IR - AMBIENT IR
                rightIRvalue = rightIRvalue - analogRead(IRright)          # REFLECTED IR = TOTAL IR - AMBIENT IR
                upIRvalue    = upIRvalue - analogRead(IRup)                # REFLECTED IR = TOTAL IR - AMBIENT IR
                downIRvalue  = downIRvalue - analogRead(IRdown)            # REFLECTED IR = TOTAL IR - AMBIENT IR
                #
                distance = (leftIRvalue + rightIRvalue + upIRvalue + downIRvalue) / 4 # distance of object is mean of reflected IR

                # ::::::::::::::
                # IRtrack.ino
                # ::::::::::::::
                # IRtrack()                                                # track objects with head
                #def IRtrack(): # =============================================== Track object in range ==========================
                # global pan
                # global tilt
                if distance < distancemax:                          #  nothing in range - return head to center position
                    if pan>PANcenter:
                        pan = pan - 5
                    if pan<PANcenter:
                        pan = pan + 5
                    if tilt>TILcenter:
                        tilt = tilt - 3
                    if tilt < TILcenter:
                        tilt = tilt + 3
                else:
                    # ---------------------------------------------------------- Track object with head --------------------
                    # Consider meaning of "/" in embedded context
                    panscale = (leftIRvalue + rightIRvalue) / LRscalefactor  # panscale & tiltscale used to adjust servo response according to distance
                    tiltscale = (upIRvalue + downIRvalue) / UDscalefactor    # adjust LRscalefactor and UDscalefactor to tweak servo responsiveness
                    #
                    leftright = (rightIRvalue - leftIRvalue) * 5 / panscale
                    pan = pan + leftright
                    updown = (downIRvalue - upIRvalue) * 5 / tiltscale
                    tilt = tilt + updown
                    # 
                    if pan < LRmin: pan=LRmin
                    if pan > LRmax: pan=LRmax
                    if tilt < UDmin: tilt=UDmin
                    if tilt > UDmax: tilt=UDmax
                #
                # 
                # PANservo.writeMicroseconds(pan)
                # TILservo.writeMicroseconds(tilt)
                PANservo.writeMicroseconds(1500-(pan-1500))                 #  reverses direction of pan/tilt servos
                TILservo.writeMicroseconds(1500-(tilt-1500))


                if distance > distancemax :
                    trick=millis()                       # reset the timer as long as object in range

            #------------------------------------------------------------ return body to standing on 4 legs --------------------
            RLKservo.writeMicroseconds(RLKcenter-300)                  # straighten rear knees slightly to lower body back down
            RRKservo.writeMicroseconds(RRKcenter-300)
            FLKservo.writeMicroseconds(FLKcenter)                      # straighten front knees to land on feet
            FRKservo.writeMicroseconds(FRKcenter)
            delay(100)                                                 # allow time for servos to respond

            RLHservo.writeMicroseconds(RLHcenter)                      # return rear hips to normal position
            RRHservo.writeMicroseconds(RRHcenter)
            delay(100)                                                 # allow time for servos to respond
            RLKservo.writeMicroseconds(RLKcenter)                      # return rear knees to normal position
            RRKservo.writeMicroseconds(RRKcenter)
            delay(100)                                                 # slow down to prevent robot falling backward

            RLKservo.writeMicroseconds(RLKcenter - 300)    # allow robot to fall back on it's feet
            RRKservo.writeMicroseconds(RRKcenter - 300)
            delay(250)
            sit = millis()                                 # reset sit timer

        if millis() - time > Speed:                        # time for next step in robot gait
            time = millis()                                # reset leg movement timer
            WalkingMotion()                                # take a step
            # ::::::::::::::
            # WalkingMotion.ino
            # ::::::::::::::
            # def WalkingMotion():
            # This function controls the leg movements to walk and turn.
            # Turning is done by make the legs on one side go foward while the other side go backward.
            # The intergers LShift and RShift determine the size and direction of the step for left and right sides.
            # Cycle controls the walking sequence.
            cycle = cycle + 1                                    # increment counter
            if cycle > 9:
                cycle=0                                          # reset counter if greater than 9

            # Note: some movements require more time than others
            #       depending on how far servos must travel.
            #       This is why 6 moves take 10 steps.
            if cycle == 0:
                FRKservo.writeMicroseconds(FRKcenter - Raise)    # raise front right leg
                RLKservo.writeMicroseconds(RLKcenter - Raise)    # raise rear  left  leg
                FLHservo.writeMicroseconds(FLHcenter - LShift)   # move  front left  leg forward / backward
                RRHservo.writeMicroseconds(RRHcenter + RShift)   # move  rear  right leg forward / backward

            elif cycle == 1:
                FRHservo.writeMicroseconds(FRHcenter - RShift)   # move  front right leg forward / backward
                RLHservo.writeMicroseconds(RLHcenter + LShift)   # move  rear  left  leg forward / backward

            elif cycle == 3:
                FRKservo.writeMicroseconds(FRKcenter)            # lower front right leg
                RLKservo.writeMicroseconds(RLKcenter)            # lower rear  left  leg

            elif cycle == 5:
                FLKservo.writeMicroseconds(FLKcenter - Raise)    # raise front left  leg
                RRKservo.writeMicroseconds(RRKcenter - Raise)    # raise rear  right leg
                FRHservo.writeMicroseconds(FRHcenter + RShift)   # move  front right leg forward / backward
                RLHservo.writeMicroseconds(RLHcenter - LShift)   # move  rear  left  leg forward / backward

            elif cycle == 6:
                FLHservo.writeMicroseconds(FLHcenter+LShift)     # move  front left  leg forward / backward
                RRHservo.writeMicroseconds(RRHcenter-RShift)     # move  rear  right leg forward / backward

            elif cycle == 8:
                FLKservo.writeMicroseconds(FLKcenter)            # lower front left  leg
                RRKservo.writeMicroseconds(RRKcenter)            #  lower rear  right leg

    else:                              # puppy is bored
        # SitShakeLiedown()                             # sits down, shakes hands or lies down
        # 
        # ::::::::::::::
        # SitShakeLiedown.ino
        #  ::::::::::::::
        # def SitShakeLiedown():
            temp = millis() - sit                     # time since last walked
            if temp < Boredom + 800:                  # Puppy in the process of sitting down
                #SittingMotion()                       # make puppy sit down slowly
                # ::::::::::::::
                # SittingMotion.ino
                # ::::::::::::::
                # def SittingMotion():
                # ---------------------------------------------------------- Sit Down -------------------------------------
                cycle = 9
                temp = (temp - Boredom) * 4                        # controls speed of sitting motion
                if temp > 3000:
                   temp = 3000

                FLHservo.writeMicroseconds(FLHcenter)              # center front hip servos
                FRHservo.writeMicroseconds(FRHcenter)

                RLHservo.writeMicroseconds(RLHcenter + (temp/6))   # set rear hips forward
                RRHservo.writeMicroseconds(RRHcenter - (temp/6))

                FLKservo.writeMicroseconds(FLKcenter)              # center front knees
                FRKservo.writeMicroseconds(FRKcenter)

                RLKservo.writeMicroseconds(RLKcenter - (temp/7))   # straighten rear knees
                RRKservo.writeMicroseconds(RRKcenter - (temp/7))

            if tilt < 800:
                sit=millis()                            # reset sit timer if puppy looking straight up
                RLHservo.writeMicroseconds(RLHcenter)   # straighten rear hips in preparation to get up
                RRHservo.writeMicroseconds(RRHcenter)
                delay(150)                              # slow down standing process to prevent robot falling backward

            else:
                # ---------------------------------------------------------- Shake Hands / Lie Down -------------------------------------------------
                if tilt < 1700:
                    trick = millis()                              # reset timer if looking up

                temp = millis() - trick                           # time how long head is looking down

                # --------------------------------------------- hand position determines trick - head follows hand - if head is central then robot will lie down
                if temp > 600:                                       # if head is looking down for more than 600mS

                    temp = (temp - 1000) * 4                          # use timer to control speed of movement
                    if temp > 3000:
                        temp=3000

                    if pan > PANcenter - 150:                            # if head is looking to the right, shake right paw
                        FRHservo.writeMicroseconds(FRHcenter - (temp/6))    # move front right hip forward
                        FRKservo.writeMicroseconds(FRKcenter - (temp/5))    # straighten front right knee

                    if pan < PANcenter + 150:                               # if head is looking to the left, shake left paw
                        FLHservo.writeMicroseconds(FLHcenter + (temp/6))    # move front left  hip forward
                        FLKservo.writeMicroseconds(FLKcenter - (temp/5))    # straighten front right knee



#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# NOTE: DO NOT EDIT THIS FILE. 
# NOTE: EDIT site/setup.tmpl instead
#

from distutils.core import setup
from distutils.version import LooseVersion
import os

def is_package(path):
    return (
        os.path.isdir(path) and
        os.path.isfile(os.path.join(path, '__init__.py'))
        )

def find_packages(path, base="" ):
    """ Find all packages in path """
    packages = {}
    for item in os.listdir(path):
        dir = os.path.join(path, item)
        if os.path.isdir(dir):
            if base:
                module_name = "%(base)s.%(item)s" % vars()
            else:
                module_name = item
            if is_package( dir ):
                packages[module_name] = dir
            packages.update(find_packages(dir, module_name))
    return packages


packages = find_packages(".")
package_names = packages.keys()

setup(name = "pyxie",
      version = "0.0.18",
      description = "Little Python to C++ Compiler",
      url='http://www.sparkslabs.com/pyxie/',
      author='Michael Sparks (sparkslabs)',
      author_email="sparks.m@gmail.com",
      license='Apache Software License',

      scripts = [
                  'bin/pyxie',
                ],

      packages = package_names,
      package_dir = packages,
      package_data={},
      long_description = """
Pyxie -- A Little Python to C++ Compiler
----------------------------------------

What job does / will this do?
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The aim of this project is to allow a adults and children to write code
in a familiar high level language that can then be compiled to run on an
arbitrary embedded system - that is devices with very low power CPUs and
very little memory. (ie devices too small to host a python
interpreter/runtime)

It's pre-alpha at the moment.

It will be useful in supporting things like the The Scout Association's
"Digital Maker" badge. That's a fair way off though.

Show me something you CAN compile
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Currently it can compile very very simple types of python program that
looks like this into an equivalent (simple) C++ program.

Example program:

::

    age = 10
    new_age = 10 +1
    new_age_too = age + 1
    new_age_three = age + new_age_too
    foo = "Hello"
    bar = "World"
    foobar = foo + bar

    print 10-1-2,7
    print 1+2*3*4-5/7,25
    print age, new_age, new_age_too
    print foo, bar, foobar

    countdown = 2147483647
    print "COUNTING DOWN"
    while countdown:
        countdown = countdown - 1

    print "BLASTOFF"

Example results:

::

    #include <iostream>
    #include <string>

    using namespace std;

    int main(int argc, char *argv[])
    {
        int age;
        string bar;
        int countdown;
        string foo;
        string foobar;
        int new_age;
        int new_age_three;
        int new_age_too;

        age = 10;
        new_age = (10+1);
        new_age_too = (age+1);
        new_age_three = (age+new_age_too);
        foo = "Hello";
        bar = "World";
        foobar = (foo+bar);
        cout << ((10-1)-2) << " " << 7 << endl;
        cout << ((1+((2*3)*4))-(5/7)) << " " << 25 << endl;
        cout << age << " " << new_age << " " << new_age_too << endl;
        cout << foo << " " << bar << " " << foobar << endl;
        countdown = 2147483647;
        cout << "COUNTING DOWN" << endl;
        while(countdown) {
            countdown = (countdown-1);
        };
        cout << "BLASTOFF" << endl;
        return 0;
    }

What does it do?
~~~~~~~~~~~~~~~~

Currently:

-  Recognise simple sequential python programs with simple statements
-  Can handle basic conditionals and while loops
-  Custom includes, and function calls to C/C++ functions (within
   limits)
-  Parse those to an AST
-  Can represent equivalent C programs using a concrete C representation
   (CST)
-  Can translate the AST to the CST and then generate C++ code from the
   CST

Python structural things it supports:

-  While loop statements
-  Comparisons
-  If/elif/elif/else statements
-  For loops - specific for X in range(Y)
-  Function calls into libraries that we link with

This is close to allowing actually useful programs now.

It's a starting point, not the end point. For that, take a look at the
language spec.

Pyxie is intended to be a simple Python to C++ compiler, with a target
of compiling python code such that it can run on a microcontroller -
like Arduino, MSP430 or ARM mbed type devices.

The name is a play on words. Specifically, Python to C++ - can be py2cc
or pycc. If you try pronouncing "pycc" it can be "pic", "py cc" or
pyc-c". The final one leads to Pixie.

This is unlikely to ever be a completely general python to C++ compiler
- if you're after than look at Shed Skin, or things like Cython, Pyrex,
and PyPy. (in terms of diminishing similarity) The difference this
project has from those is that this project assumes a very small target
device. Something along the lines of an Atmega 8A, Atmega 328 or more
capable.

In the past I've written a test driven compiler suite, so I'll be
following the same approach here. It did consider actually making Pyxie
use that as a frontend, but for the moment, I'd like python
compatibility.

Why not micropython? Micropython is **ace** . If your device is large
enough to support the micropython runtime, use it! The aim of this is on
the really small microcontrollers- the ones too small to even support
micropython - like an MSP430, or an Atmega 8A or similarly tiny MCU.

In the past I've written a test driven compiler suite, so I'll be
following the same approach here. It did consider actually making Pyxie
use that as a frontend, but for the moment, I'd like python
compatibility.

Status Overview
---------------

For the impatient: this probably does **NOT** do what you want, **yet**.

High level view of support:

-  Supports variables, sequence, and assignment
-  while loops controlled by expressions, possibly involving variables
-  while loops can contain break/continue which allows "if" style
   functionality
-  Also have basic conditional operators like "==", "!=", etc.
-  Ability to pull in C++ includes on standard paths

This means we can almost start writing useful programs, but in
particular can start creating simplistic benchmarks for measuring run
speed. It IS getting there however, and feedback, usecases, devices very
welcome.

Influences
----------

Many moons ago, I made a generic language parser which I called SWP
(semantic whitespace parser), or Gloop.

-  https://github.com/sparkslabs/minisnips/tree/master/SWP
-  http://www.slideshare.net/kamaelian/swp-a-generic-language-parser

It was an experiment to see if you could write a parser that had no
keywords, or similar, in a completely test driven fashion. ie a bit like
a parser for a Lisp like language that would look like python or ruby.
It turns out that you can and there's lots of interesting things that
arise if you do. (Best seen in the slideshare link)

Which version of Python?
------------------------

It's not a complete subset of any particular python, but it's based
around the intersection points in python 2 and 3. It will be, by
definition, a non-dynamic subset - at least at first.

-  For detail as to what's planned for the language, take a look at the
   language spec.
-  For an overview as to the guiding principles, please take a look at
   project status
-  For detail as to what's actually implemented, take a look at language
   status

These are all a WIP, but becoming more solid.

Why write this?
---------------

Personally, having built something simpler in the past, I know I'd find
it useful. (I use python rather than C++ often because I can write more
quicker with the former). Also, I work with kids in my spare time, and
it opens up options there.

I've written something like this for work last year, but that was much
more limited and restricted in both aspiration and implementation. This
rewrite is something I've done on my own time, with my own tools, from
scratch, which allows me to share this with others.

Major changes:

-  This aims to be a more rounded implementation
-  This performs transforms from an AST (abstract syntax tree) to a CCR
   (concrete code representation), rather than munging code directly
   from a concrete parse tree.

That potentially allows other things, like creation of visual
representations of programs from code as well.

Is this part of any larger project?
-----------------------------------

No. It could be used by others, but it's got a definite goal - to allow
the use of a "little" python to program devices which are too small to
host a python runtime.

If anything, it's a continuation of the personal itch around SWP
(mentioned above) from about 10 years ago. Unlike that though, it's
much, much better structured.

One thing that may happen though is the ability to take python classes
and derive iotoy device implementations/interfaces directly. (since
iotoy was inspired heavily by python introspection) That's quite some
time off.

Release History
---------------

Release History:

-  0.0.18 - 2016-01-10 - Grammar changes to support object attributes
   and methods, start of servo support in arduino profile.
-  0.0.17 - 2015-08-12 - Add pass statement, enable "for" on arduino,
   update documentation, refactor pyxie harness
-  0.0.16 - 2015-08-02 - Adds initial Arduino LEONARDO support, improved
   function call, release build scripts
-  0.0.15 - 2015-07-18 - clib converted to py clib for adding to build
   directory
-  0.0.14 - 2015-07-18 - For loops implemented. Added clib code, C++
   generator implementation, FOR loop style test harness, parsing and
   basic analysis of of FOR loops using a range interator
-  0.0.13 - 2015-06-21 - if/elif/else,conditionals/boolean/parenthesised
   expressions.
-  0.0.12 - 2015-06-16 - While loops, break/continue, Website,
   comparison operators, simple benchmark test
-  0.0.11 - 2015-06-06 - Function calls; inclusion of custom C++
   headers; empty statements; language spec updates
-  0.0.10 - 2015-06-03 - Analysis phase to make type inference work
   better. Lots of related changes. Implementation of expression
   statements.
-  0.0.9 - 2015-05-23 - Grammar changed to be left, not right recursive.
   (Fixes precedence in un-bracketed expressions) Added standalone
   compilation mode - outputs binaries from python code.
-  0.0.8 - 2015-05-13 - Internally switch over to using node objects for
   structure - resulting in better parsing of expressions with variables
   and better type inference.
-  0.0.7 - 2015-04-29 - Structural, testing improvements, infix
   operators expressions (+ - \* / ) for integers, precdence fixes
-  0.0.6 - 2015-04-26 - Character Literals, "plus" expressions,
   build/test improvements
-  0.0.5 - 2015-04-23 - Core lexical analysis now matches language spec,
   including blocks
-  0.0.4 - 2015-04-22 - Mixed literals in print statements
-  0.0.3 - 2015-04-21 - Ability to print & work with a small number of
   variables
-  0.0.2 - 2015-03-30 - supports basic assignment
-  0.0.1 - Unreleased - rolled into 0.0.2 - Initial structure

Language Status
---------------

::

    program : statements
    statements : statement
               | statement statements

    statement_block : INDENT statements DEDENT

    statement : assignment_statement
              | print_statement
              | general_expression
              | EOL
              | while_statement
              | break_statement
              | continue_statement
              | pass_statement
              | if_statement
              | for_statement

    assignment_statement -> IDENTIFIER ASSIGN general_expression # ASSIGN is currently limited to "="

    while_statement : WHILE general_expression COLON EOL statement_block

    break_statement : BREAK

    pass_statement : PASS

    continue_statement : CONTINUE

    if_statement : IF general_expression COLON EOL statement_block
                 | IF general_expression COLON EOL statement_block extended_if_clauses

    extended_if_clauses : else_clause
                        | elif_clause

    else_clause : ELSE COLON EOL statement_block

    elif_clause : ELIF general_expression COLON EOL statement_block
                | ELIF general_expression COLON EOL statement_block extended_if_clauses

    print_statement : 'print' expr_list # Temporary - to be replaced by python 3 style function

    for_statement | FOR IDENTIFIER IN general_expression COLON EOL statement_block

    expr_list : general_expression
              | general_expression COMMA expr_list

    general_expression : boolean_expression

    boolean_expression : boolean_and_expression
                       | boolean_expression OR boolean_and_expression

    boolean_and_expression : boolean_not_expression
                           | boolean_and_expression AND boolean_not_expression

    boolean_not_expression : relational_expression
                           | NOT boolean_not_expression

    relational_expression : expression
                          | relational_expression COMPARISON_OPERATOR expression

    expression : arith_expression
               | expression '+' arith_expression
               | expression '-' arith_expression
               | expression '**' arith_expression

    arith_expression : negatable_expression_atom
                     | arith_expression '*' negatable_expression_atom
                     | arith_expression '/' negatable_expression_atom


    negatable_expression_atom : "-" negatable_expression_atom 
                              | expression_atom

    expression_atom : value_literal
                    | IDENTIFIER '(' ')' # Function call, with no arguments
                    | IDENTIFIER '(' expr_list ')' # Function call
                    | '(' general_expression ')'

    value_literal : number
                  | STRING
                  | CHARACTER
                  | BOOLEAN
                  | IDENTIFIER

    number : NUMBER
           | FLOAT
           | HEX
           | OCTAL
           | BINARY
           | LONG         (suffice is L)
           | UNSIGNEDLONG (suffice is l)
           | '-' number

Current Lexing rules used by the grammar:

::

    NUMBER : \d+
    FLOAT : \d+.\d+ # different from normal python, which allows .1 and 1.
    HEX : 0x([abcdef]|\d)+
    OCTAL : 0o\d+
    BINARY : 0b\d+
    STRING - "([^\"]|\.)*" or '([^\']|\.)*' # single/double quote strings, with escaped values
    CHARACTER : c'.' /  c"." # Simplification - can be an escaped character
    BOOLEAN : True|False
    IDENTIFIER : [a-zA-Z_][a-zA-Z0-9_]*

Limitations
-----------

Most expressions currently rely on the C++ counterparts. As a result not
all combinations which are valid are directly supported yet. Notable
ones:

-  Combinations of strings with other strings (outlawing /\*, etc)
-  Combinations of strings with numbers

Why a python 2 print statement?
-------------------------------

Python 2 has print statement with special notation; python 3's version
is a function call. The reason why this grammar currently has a python-2
style print statement with special notation is to specifically avoid
implementing general function calls yet. Once those are implemented,
special cases - like implementing print - can be implemented, and this
python 2 style print statement WILL be removed. I expect this will occur
around version 0.0.15, based on current rate of progress.

Keeping it for now also simplifies "yield" later

Michael Sparks, January 2015

"""
      )



#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import pprint
import time

import pyxie.codegen.profiles as profiles
import pyxie.codegen.simple_cpp

from pyxie.parsing.lexer import build_lexer
from pyxie.parsing.grammar import parse

from pyxie.model.pynode import jdump
from pyxie.model.transform import ast_to_cst

from pyxie.codegen.clib import files as clib_files
from pyxie.codegen.simple_cpp import C_Program, source, reset_parser

testdir = "test-data"
testprogs_dir = os.path.join(testdir, "progs")


def remove_directory(build_dir):
    for filename in os.listdir(build_dir):
        try:
            os.unlink(os.path.join(build_dir, filename))
        except OSError, e:
            if e.errno == 21:
                remove_directory(os.path.join(build_dir, filename))
    os.rmdir(build_dir)


def get_test_programs(program_suffix):
    testprogs = [x for x in os.listdir(testprogs_dir) ]
    testprogs = [x for x in testprogs if x.endswith(program_suffix) ]
    return testprogs


def get_build_environment(filename, result_filename):
    rootdir = os.getcwd()

    lastslash_pos = filename.rfind("/")
    if lastslash_pos != -1:
        base_dir = filename[:lastslash_pos]
        base_filename = filename[lastslash_pos+1:]
    else:
        base_dir = "." # just a filename
        base_filename = filename

    cname = base_filename[:base_filename.rfind(".")]

    if not result_filename:
        result_filename = os.path.join(base_dir, cname)

    return {
            "rootdir" :rootdir,
            "base_dir" : base_dir,
            "base_filename":base_filename,
            "cname" : cname,
            "result_filename" : result_filename,
           }


def parse_file(somefile):
    lexer = build_lexer()
    reset_parser()
    data = open(somefile).read() + "\n#\n"
    AST = parse(data, lexer)
    AST.includes = lexer.includes
    print "AST includes", AST.includes
    print "PRINT AST ----------------------------------"
    print AST
    print "---------------------------------- PRINT AST"
    return AST


def analyse_file(filename):
    try:
        AST = parse_file(filename)
        print "HERE 3------------"
        pprint.pprint(AST.__json__())
        print "HERE 4------------"
        AST.analyse()
        print "HERE 5------------"
        pprint.pprint(AST.__info__())
        print "HERE 6------------"
    except:
        from pyxie.parsing.context import Context
        for cid in Context.contexts:
            context = Context.contexts[cid]
            print "CONTEXT",context.names
        raise


def generate_code(cname, AST, profile, debug=False):
    CST = ast_to_cst(cname, AST)
    if debug:
        print "CONCRETE C SYNTAX TREE: - - - - - - - - - - - - - - - - - - - -"
        print pprint.pformat(CST)
        print "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -"
    program = C_Program.fromjson(CST)
    program.generate(profile)
    return pyxie.codegen.simple_cpp.source[:]


def codegen_phase(filename, result_filename, profile):
    build_env = get_build_environment(filename, result_filename)

    cname = build_env["cname"]

    AST = parse_file(filename)
    AST.analyse()
    print "_______________________________________________________________"
    print "GENERATING CODE"
    print
    c_code = generate_code(cname, AST, profile, debug=True)  # Need the C NAME TO DO THIS
    print "_______________________________________________________________"
    print "Generated Program"
    print "\n".join(c_code)
    print "_______________________________________________________________"
    return c_code


def build_program(source, work_dir, name, profile):
    print "_______________________________________________________________"
    print "BUILDING", work_dir
    print
    print "Program to build:"
    pprint.pprint(source, width=200)
    print

    extension = profiles.mainfile_extensions.get(profile,"c")

    f = open(os.path.join(work_dir,name+"." +extension), "w")
    for line in source:
        f.write(line)
        f.write("\n")
    f.close()


    try:
        makefile_tmpl = profiles.makefile_templates[profile]
    except KeyError:
        makefile_tmpl = profiles.makefile_templates["default"]

    makefile = makefile_tmpl % {"filename": name }
    f = open(os.path.join(work_dir,"Makefile"), "w")
    f.write(makefile)
    f.close()

    exclusions = profiles.clib_exclusions.get(profile, [])
    for filename in clib_files:
        if filename.endswith("_test.cpp"):
            # Skip main programs to avoid confusing profiles
            continue
        if filename in exclusions:
            print "Skipping", filename, "for profile", profile
            continue

        f = open( os.path.join(work_dir,filename), "w")
        f.write(clib_files[filename])
        f.close()

    os.chdir(work_dir)
    os.system("make")
    print
    print "Done!"
    print


# Next function is called by compile_file, but could be called independently
# This is why both do "build_env", rather that expect the caller to
def compile_file(filename, profile, result_filename=None):

    build_env = get_build_environment(filename, result_filename)

    rootdir = build_env["rootdir"]
    base_dir = build_env["base_dir"]
    base_filename = build_env["base_filename"]
    cname = build_env["cname"]
    result_filename = build_env["result_filename"]

    c_code  = codegen_phase(filename, result_filename, profile)

    print "COMPILING", filename
    print "IN", base_dir
    print "SOURCEFILE", base_filename
    print "cname", cname
    print "result_filename", result_filename

    build_dir = os.path.join(base_dir, "build-"+str(int(time.time())))
    try:
        os.mkdir(build_dir)
    except OSError as e:
       if e.errno != 17: # Directory exists
           raise

    build_program(c_code, build_dir, cname, profile)

    os.chdir(rootdir)

    actual_result_file = (profiles.result_file[profile])(build_dir, cname)

    result_filename = (profiles.modify_result_file[profile])(result_filename)

    print "BUILD DIR", build_dir
    print "RESULT FILE", actual_result_file
    print "DEST FILE", result_filename
    print "CWD", os.getcwd()

    os.rename(actual_result_file,result_filename)
    clean = False
    if clean:
        remove_directory(build_dir)
        if base_dir == ".":
            os.unlink("parsetab.py")
            os.unlink("parser.out")
    return


def parse_testfile(testprogs_dir, testfile, debug=False):
    print "_______________________________________________________________"
    print "PARSING", os.path.join(testprogs_dir,testfile)
    print
    AST = parse_file(os.path.join(testprogs_dir,testfile))
    if debug:
        pprint.pprint(AST)
        JAST = jdump(AST)
        pprint.pprint(JAST)
    return AST


def compile_testfile(testprogs_dir, testfile, profile):
    # The compiled c name is the filename with the suffix removed
    p = testfile.rfind(".")
    cname = testfile[:p]

    AST = parse_testfile(testprogs_dir, testfile, debug=True)

    print "_______________________________________________________________"
    print "ANALYSING", os.path.join(testprogs_dir,testfile)
    print
    AST.analyse()

    print "_______________________________________________________________"
    print "COMPILING", os.path.join(testprogs_dir,testfile)
    print

    c_code = generate_code(cname, AST, profile, debug=True)

    # Create Work directory
    allresults_dir = os.path.join(testdir, "genprogs")
    thisresult_dir = os.path.join(allresults_dir, testfile)

    try:
        os.mkdir(allresults_dir)
    except OSError as e:
        pass # Can raise an exception if the directory exists, which is fine

    try:
        os.mkdir(thisresult_dir)
    except OSError as e:
       if e.errno != 17: # Directory exists
           raise

    build_program(c_code, thisresult_dir, cname, profile)


def parsing_tests():
    # Run parsing tests. These are in the "progs" test directory, and are in
    # filenames ending ".p"

    rootdir = os.getcwd()
    testprogs = get_test_programs(".p")

    for testfile in testprogs:
        AST = parse_testfile(testprogs_dir, testfile)
        pprint.pprint(AST)

        JAST = jdump(AST)
        pprint.pprint(JAST)

        os.chdir(rootdir)


def compilation_tests(profile):
    # ast_to_cst
    # Compilation Tests
    rootdir = os.getcwd()
    testprogs = get_test_programs(".pyxie")
    print "TEST PROGS", testprogs

    for testfile in testprogs:
        build_fail = True
        try:
            compile_testfile(testprogs_dir, testfile, profile)
            build_fail = False
        except:
            if testfile.endswith("shouldfail.pyxie"):
                print "AS EXPECTED, COMPILE FAILED FOR", testfile
            else:
                print "UNEXPECTED FAILURE FOR FILE", testfile
                raise

        if not build_fail and testfile.endswith("shouldfail.pyxie"):
            print "UNEXPECTED BUILD SUCCESS FOR FILE", testfile
            raise Exception("UNEXPECTED BUILD SUCCESS FOR FILE %s" % testfile )

        os.chdir(rootdir)

    print "COMPILING DONE", testprogs
